/*
 * PPF c++ wrapper
 *
 * Copyright (C) 2008 Alex Meakins
 *
 * 2008-04-09: Alex Meakins <alex.meakins@jet.uk>
 *             initial version
 *
 * This file is part of PPFc++.
 *
 * PPFc++ is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * PPFc++ is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser Public License for more details.
 *
 * You should have received a copy of the GNU Lesser Public License
 * along with PPFc++.  If not, see <http://www.gnu.org/licenses/>.
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#ifndef __PPFDATATYPE_HPP__
#define __PPFDATATYPE_HPP__

#include <vector>
#include <string>
#include "ppftypes.hpp"

using namespace std;

namespace ppf
{
	class system;

	/// Contains signal data requested from the PPF system.
	class datatype
	{

		// friend functions ----------------------------------------------------

			/// Datatype is a friend of system.
			friend class system;

		public:

		// constructors/destructors

			// basic constructors ----------------------------------------------

			/// Constructor
			datatype() : xunit("        "), tunit("        "), dataunit("        "), comment("                        "), userstatus(0), sysstatus(0) { };

			// create constructors ---------------------------------------------

//			datatype(const int nt, const int nx) : xunit(""), tunit(""), dataunit(""), comment(""), userstatus(0), sysstatus(0) { resizeTX(nt, 0.0, 0.0, nx, 0.0, 0.0); };
//			datatype(const int nt, const float t, const float dt, const int nx, const float x, const float dx) : xunit(""), tunit(""), dataunit(""), comment(""), userstatus(0), sysstatus(0) { createTX(nt, t, dt, nx, x, dx); };

			// read constructors -----------------------------------------------

			/// Initialising constructor.
			/// Reads in data from a local file.
			/// \param filename String containing the filename (including path)
			datatype(const string filename) : xunit("        "), tunit("        "), dataunit("        "), comment("                        "), userstatus(0), sysstatus(0) { this->readLocal(filename); };

		// class methods -------------------------------------------------------

			// resize ----------------------------------------------------------

			/// Resizes the T and data vectors to the specified number of elements along the T axis.
			/// Any contents in the data and T vectors at indicies outside the T range will be lost.
			/// If the datatype vectors are initially zero sized and the new T size is > 0 the data vector will be expanded along the X axis by a single element to provide space for 1D data.
			/// Setting the T size to zero will destroy the contents of the T, X and data vectors.
			/// \param nt Number of T elements.
			/// \throws ppf_error If an invalid number of elements is passed.
			inline void resizeT(const int nt) { resizeTX(nt, (xvec.empty() ? 1 : xvec.size())); };

			/// Resizes the X and data vectors to the specified number of elements along the X axis.
			/// Any contents in the data and X vectors at indicies outside the X range will be lost.
			/// If the datatype vectors are initially zero sized and the new X size is > 0 the data vector will be expanded along the T axis by a single element to provide space for 1D data.
			/// Setting the X size to zero will destroy the contents of the T, X and data vectors.
			/// \param nx Number of X elements.
			/// \throws ppf_error If an invalid number of elements is passed.
			inline void resizeX(const int nx) { resizeTX((tvec.empty() ? 1 : tvec.size()), nx); };

			/// Resizes the T, X and data vectors to the specified number of elements along the T and X axes.
			/// Any contents at indicies outside the new T and X ranges will be lost.
			/// Setting either the X or T sizes are set to zero the contents of T, X and data vectors will be destroyed.
			/// \param nt Number of T elements.
			/// \param nx Number of X elements.
			/// \throws ppf_error If an invalid number of elements is passed.
			void resizeTX(int nt, int nx);

			/// Clears the datatype by setting it back to its default values.
			/// All contents of the datatype will be erased including comments, units and status flags.
			void clear();

			/// Returns true if the T and X dimensions of the datatype are zero, false otherwise.
			/// Meta-data such as the comments, units and status flags are not checked by a call to empty().
			inline bool empty() const { return (tvec.size() || xvec.size()) ? false : true; };

			// read ------------------------------------------------------------

			// read from local

			/// Reads in data from a local file.
			/// \param filename String containing the filename (including path)
			void readLocal(const string filename);

			// write -----------------------------------------------------------

			// write to local

			/// Reads in data from a local file.
			/// \param filename String containing the filename (including path)
			void writeLocal(const string filename);

			// get -------------------------------------------------------------

			// get datatype meta data

			/// Returns the T dimension of the data and T vectors
			/// \return The T dimension of the data and T vectors
			inline int sizeT() const { return tvec.size(); };

			/// Returns the X dimension of the data and X vectors
			/// \return The X dimension of the data and X vectors
			inline int sizeX() const { return xvec.size(); };

			/// Returns the units of the T vector
			/// \return A string containing the units of the T vector
			inline string getTUnit() const { return tunit; };

			/// Returns the units of the X vector
			/// \return A string containing the units of the X vector
			inline string getXUnit() const { return xunit; };

			/// Returns the units of the data vector
			/// \return A string containing the units of the data vector
			inline string getDataUnit() const { return dataunit; };

			/// Returns the datatype comment
			/// \return A string containing the datatype comment
			inline string getComment() const { return comment; };

			/// Returns the user status of the datatype
			/// \return An integer containing the status value
			inline int getUserStatus() const { return userstatus; };

			/// Returns the system status of the datatype
			/// \return An integer containing the status value
			inline int getSysStatus() const { return sysstatus; };

			// get data by index

			/// Returns the T vector value at index \e It
			/// \param It index in T-direction
			/// \return The T vector value at index \e It
			inline float getT(const int It) const { return tvec[It]; };

			/// Returns the X vector value at index \e Ix
			/// \param Ix index in X-direction
			/// \return The X vector value at index \e Ix
			inline float getX(const int Ix) const { return xvec[Ix]; };

			/// Returns the data value at index \e It (\e Ix=0).
			/// If the ppf is 2D then the X index is always taken to be zero.
			/// \param It index in T-direction
			/// \return The data value at index \e It
			inline float getDataIt(const int It) const { return data[It][0]; };

			/// Returns the data value at index \e Ix (\e It=0).
			/// If the ppf is 2D then the T index is always taken to be zero.
			/// \param Ix index in X-direction
			/// \return The data value at index \e Ix
			inline float getDataIx(const int Ix) const { return data[0][Ix]; };

			/// Returns the data value at index [\e It, \e Ix].
			/// \param It index in T-direction
			/// \param Ix index in X-direction
			/// \return The data value at index [\e It, \e Ix].
			inline float getDataItIx(const int It, const int Ix) const { return data[It][Ix]; };

			// get data by co-ordinate

			/// Returns the closest T vector index to the passed time \e t.
			/// This routine will only be successful provided the T vector is monotonically increasing or decreasing and no two values are equal, its result is undefined if applied to data not satisfying these conditions.
			/// \param t time
			/// \return The T vector index closest to passed time \e t
			int getIt(const float t) const;

			/// Returns the closest X vector index to the passed \e x co-ordinate.
			/// This routine will only be successful provided the X vector is monotonically increasing or decreasing and no two values are equal, its result is undefined if applied to data not satisfying these conditions.
			/// \param x co-ordinate
			/// \return The X vector index closest to passed \e x co-ordinate.
			int getIx(const float x) const;

			/// Returns linearly interpolated data value at time \e t (assumes \e Ix=0).
			/// This routine will only be successful provided the T vector is monotonically increasing or decreasing and no two values are equal, its result is undefined if applied to data not satisfying these conditions.
			/// \param t time
			/// \return The linearly interpolated data value at time \e t
   			inline float getDataT(const float t) const { return getDataT(t, ppf::linear); };

			/// Returns the data value at time \e t using the method specified by the passed process flag (assumes \e Ix=0).
			/// Valid process flags are: \e nearest, \e linear, \e bicubic
			/// This routine will only be successful provided the T vector is monotonically increasing or decreasing and no two values are equal, its result is undefined if applied to data not satisfying these conditions.
			/// \param t time
			/// \param flag a valid process flag
			/// \return The data value
			float getDataT(const float t, const processflag flag) const;

			/// Returns a packed coordinate (t, x, data value) at time \e t using the method specified by the passed process flag (assumes \e Ix=0).
			/// Valid process flags are: \e nearest, \e linear, \e bicubic
			/// This routine will only be successful provided the T vector is monotonically increasing or decreasing and no two values are equal, its result is undefined if applied to data not satisfying these conditions.
			/// \param t time
			/// \param flag a valid process flag
			/// \return A packed coordinate containing the exact \e t, \e x and \e data values
			coord getDataCT(const float t, const processflag flag) const;

			/// Returns linearly interpolated data value at \e x (assumes \e It=0).
			/// This routine will only be successful provided the X vector is monotonically increasing or decreasing and no two values are equal, its result is undefined if applied to data not satisfying these conditions.
			/// \param x x co-ordinate
			/// \return The linearly interpolated data value at \e x
			inline float getDataX(const float x) const { return getDataX(x, ppf::linear); };

			/// Returns the data value at \e x using the method specified by the passed process flag (assumes \e It=0).
			/// Valid process flags are: \e nearest, \e linear, \e bicubic
			/// This routine will only be successful provided the X vector is monotonically increasing or decreasing and no two values are equal, its result is undefined if applied to data not satisfying these conditions.
			/// \param x x co-ordinate
			/// \param flag a valid process flag
			/// \return The data value
			float getDataX(const float x, const processflag flag) const;

			/// Returns a packed coordinate (t, x, data value) at \e x using the method specified by the passed process flag (assumes \e It=0).
			/// Valid process flags are: \e nearest, \e linear, \e bicubic
			/// This routine will only be successful provided the X vector is monotonically increasing or decreasing and no two values are equal, its result is undefined if applied to data not satisfying these conditions.
			/// \param x x co-ordinate
			/// \param flag a valid process flag
			/// \return A packed coordinate containing the exact t, x and data values
			coord getDataCX(const float x, const processflag flag) const;

			/// Returns linearly interpolated data value at time \e t and X index \e Ix.
			/// This routine will only be successful provided the T vector is monotonically increasing or decreasing and no two values are equal, its result is undefined if applied to data not satisfying these conditions.
			/// \param t time
			/// \param Ix index in X-direction
			/// \return The linearly interpolated data value at time \e t and index \e Ix
			inline float getDataTIx(const float t, const long Ix) const { return getDataTIx(t, Ix, ppf::linear); };

			/// Returns the data value at time \e t and X index \e Ix using the method specified by the passed process flag.
			/// Valid process flags are: \e nearest, \e linear, \e bicubic
			/// This routine will only be successful provided the T vector is monotonically increasing or decreasing and no two values are equal, its result is undefined if applied to data not satisfying these conditions.
			/// \param t time
			/// \param Ix index in X-direction
			/// \param t_flag a valid process flag
			/// \return The data value
			float getDataTIx(const float t, const long Ix, const processflag t_flag) const;

			/// Returns a packed coordinate (t, x, data value) at time \e t and X index \e Ix using the method specified by the passed process flag.
			/// Valid process flags are: \e nearest, \e linear, \e bicubic
			/// This routine will only be successful provided the T vector is monotonically increasing or decreasing and no two values are equal, its result is undefined if applied to data not satisfying these conditions.
			/// \param t time
			/// \param Ix index in X-direction
			/// \param t_flag a valid process flag
			/// \return A packed coordinate containing the exact \e t, \e x and \e data values
			coord getDataCTIx(const float t, const long Ix, const processflag t_flag) const;

			/// Returns linearly interpolated data value at \e x and T index \e It.
			/// This routine will only be successful provided the X vector is monotonically increasing or decreasing and no two values are equal, its result is undefined if applied to data not satisfying these conditions.
			/// \param x x co-ordinate
			/// \param It index in T-direction
			/// \return The linearly interpolated data value at \e x and index \e It
			inline float getDataItX(const long It, const float x) const { return getDataItX(It, x, ppf::linear); };

			/// Returns the data value at \e x and T index \e It using the method specified by the passed process flag.
			/// Valid process flags are: \e nearest, \e linear, \e bicubic
			/// This routine will only be successful provided the X vector is monotonically increasing or decreasing and no two values are equal, its result is undefined if applied to data not satisfying these conditions.
			/// \param x x co-ordinate
			/// \param It index in T-direction
			/// \param x_flag a valid process flag
			/// \return The data value
			float getDataItX(const long It, const float x, const processflag x_flag) const;

			/// Returns a packed coordinate (t, x, data value) at \e x and T index \e It using the method specified by the passed process flag (assumes \e It=0).
			/// Valid process flags are: \e nearest, \e linear, \e bicubic
			/// This routine will only be successful provided the X vector is monotonically increasing or decreasing and no two values are equal, its result is undefined if applied to data not satisfying these conditions.
			/// \param x x co-ordinate
			/// \param It index in T-direction
			/// \param x_flag a valid process flag
			/// \return A packed coordinate containing the exact t, x and data values
			coord getDataCItX(const long It, const float x, const processflag x_flag) const;

			/// Returns linearly interpolated data value at time \e t and co-ordinate \e x.
			/// This routine will only be successful provided the T and X vectors are monotonically increasing or decreasing and no two values are equal in each vector, its result is undefined if applied to data not satisfying these conditions.
			/// \param t time
			/// \param x x co-ordinate
			/// \return The linearly interpolated data value at time \e t and co-ordinate \e x
			inline float getDataTX(const float t, const float x) const { return getDataTX(t, x, ppf::linear, ppf::linear); };

			/// Returns the data value at time \e t and co-ordinate \e x using the method specified by the passed process flag.
			/// Valid process flags are: \e nearest, \e linear, \e bicubic
			/// This routine will only be successful provided the T and X vectors are monotonically increasing or decreasing and no two values are equal in each vector, its result is undefined if applied to data not satisfying these conditions.
			/// \param t time
			/// \param x x co-ordinate
			/// \param t_flag a valid process flag for the T vector
			/// \param x_flag a valid process flag for the X vector
			/// \return The data value
			float getDataTX(const float t, const float x, const processflag t_flag, const processflag x_flag) const;

			/// Returns a packed coordinate (t, x, data value) at time \e t and co-ordinate \e x using the method specified by the passed process flag.
			/// Valid process flags are: \e nearest, \e linear, \e bicubic
			/// This routine will only be successful provided the T and X vectors are monotonically increasing or decreasing and no two values are equal in each vector, its result is undefined if applied to data not satisfying these conditions.
			/// \param t time
			/// \param x x co-ordinate
			/// \param t_flag a valid process flag for the T vector
			/// \param x_flag a valid process flag for the X vector
			/// \return A packed coordinate containing the exact \e t, \e x and \e data values
			coord getDataCTX(const float t, const float x, const processflag t_flag, const processflag x_flag) const;

			// set -------------------------------------------------------------

			// set datatype meta data

			/// Sets the units of the T vector.
			/// The maximum string length is 8 chars, characters beyond this limit will be removed.
			/// Strings shorter than 8 chars will be automatically padded with spaces.
			/// \param s T vector unit string.
			void setTUnit(const string s);

			/// Sets the units of the X vector
			/// The maximum string length is 8 chars, characters beyond this limit will be removed.
			/// Strings shorter than 8 chars will be automatically padded with spaces.			
			/// \param s X vector unit string.
			void setXUnit(const string s);

			/// Sets the units of the data vector
			/// The maximum string length is 8 chars, characters beyond this limit will be removed.
			/// Strings shorter than 8 chars will be automatically padded with spaces.
			/// \param s Data vector unit string.
			void setDataUnit(const string s);

			/// Sets the datatype comment
			/// The maximum string length is 24 chars, characters beyond this limit will be removed.			
			/// Strings shorter than 24 chars will be automatically padded with spaces.			
			/// \param s Comment string.
			void setComment(const string s);

			/// Sets the user status of the datatype
			/// \param v Datatype user status flag.
			inline void setUserStatus(const int v) { userstatus = v; };

			/// Sets the system status of the datatype
			/// \param v Datatype system status flag.
			inline void setSysStatus(const int v) { sysstatus = v; };

			// set data by index

			/// Sets the T vector value at index \e It
			/// \param It index in T-direction
			/// \param v new T vector value
			inline void setT(const int It, const float v) { tvec[It] = v; };

			/// Sets the X vector value at index \e Ix
			/// \param Ix index in X-direction
			/// \param v new X vector value
			inline void setX(const int Ix, const float v) { xvec[Ix] = v; };

			/// Sets the data value at index \e It (\e Ix=0).
			/// If the ppf is 2D then the X index is always taken to be zero.
			/// \param It index in T-direction
			/// \param v new data value
			inline void setDataIt(const int It, const float v) { data[It][0] = v; };

			/// Sets the data value at index \e Ix (\e It=0).
			/// If the ppf is 2D then the T index is always taken to be zero.
			/// \param Ix index in X-direction
			/// \param v new data value
			inline void setDataIx(const int Ix, const float v) { data[0][Ix] = v; };

			/// Sets the data value at index [\e It, \e Ix].
			/// \param It index in T-direction
			/// \param Ix index in X-direction
			/// \param v new data value
			inline void setDataItIx(const int It, const int Ix, const float v) { data[It][Ix] = v; };

			// set data by co-ordinate
			// set routines can take nearest and insert process flags - nearest sets the value for the nearest point, insert inserts a new element in the array and data tables and assigns the passed value

		private:

		// variables

			vector<float> tvec;				///< Holds the T vector values if required.
			vector<float> xvec;				///< Holds the X vector values if required.
			vector<vector<float> > data;	///< Holds the 1D or 2D data values.
			string xunit;                   ///< Contains the x vector units (limited to 8 chars)
			string tunit;                   ///< Contains the t vector units (limited to 8 chars)
			string dataunit;                ///< Contains the data vector units (limited to 8 chars)
			string comment;                 ///< Contains datatype comment (limited to 24 chars)
			int userstatus;					///< Holds the user datatype status value
			int sysstatus;					///< Holds the system datatype status value

		// functions

			// process ---------------------------------------------------------

			/// This routine searches the passed vector over a specified range to find either the element which exactly matches the passed value \e f or the two elements which bound \e f in value.
			/// Returns the indicies (\e Ilower and \e Iupper) of the vector elements surrounding the passed value \e f, searching is performed over the range [\e Imin, \e Imax].
			/// \e Ilower will contain the index of the element of closest value smaller than \e f.
			/// \e Iupper will contain the index of the element of closest value larger than \e f.
			/// If \e Iupper = \e Ilower then either an exact index was found corresponding to the value of \e f which is given by either \e Iupper or \e Ilower.
			/// This routine will only be successful provided the vector is monotonically increasing or decreasing, its result is undefined if applied to data not satisfying these conditions.
			/// The number order of indexes of \e Ilower and \e Iupper will depend on the order of the values in the vector. If the vector values increase with index then \e Ilower < \e Iupper, otherwise \e Ilower > \e Iupper.
			/// \param v vector to search
			/// \param Imin minimum index to search
			/// \param Imax maximum index to search
			/// \param f value of interest
			/// \param Ilower (returned) index of closest element of lower value than \e f
			/// \param Iupper (returned) index of closest element of higher value than \e f
			void findNearest(const vector<float> &v, const int Imin, const int Imax, const float f, int &Ilower, int &Iupper) const;

			/// Performs a linear interpolation for the passed parameters.
			/// \e pt must lie between p1 and p2, the operation of this routine outside those bounds is undefined.
			/// \param p1 co-ordinate point 1
			/// \param v1 value at co-ordinate 1
			/// \param p2 co-ordinate point 2
			/// \param v2 value at co-ordinate 2
			/// \param pt target point for interpolation
			/// \returns interpolated value
			inline double interpolateLinear(const double p1, const double v1, const double p2, const double v2, const double pt) const { return (v1 + ((v2-v1)/(p2-p1))*(pt-p1)); };

			/// Performs a cubic interpolation for the passed parameters.
			/// \e pt must lie between p1 and p2, the operation of this routine outside those bounds is undefined.
			/// \param p1 co-ordinate point 1
			/// \param g1 dv/dp at point 1
			/// \param v1 value at co-ordinate 1
			/// \param p2 co-ordinate point 2
			/// \param g2 dv/dp at point 2
			/// \param v2 value at co-ordinate 2
			/// \param pt target point for interpolation
			/// \returns interpolated value
			double interpolateCubic(const double p1, const double g1, const double v1, const double p2, const double g2, const double v2, const double pt) const;

			/// Common, internal functionality of time interpolating data value get methods.
			/// Valid process flags are: \e nearest, \e linear, \e bicubic
			/// This routine will only be successful provided the T vector is monotonically increasing or decreasing and no two values are equal, its result is undefined if applied to data not satisfying these conditions.
			/// \param t time
			/// \param Ix index in X-direction
			/// \param t_flag a valid process flag
			/// \param ta (returned) actual t value obtained via selected process
			/// \param v (returned) The data value
			void getDataTIx(const float t, const long Ix, const processflag t_flag, float &ta, float &v) const;

			/// Common, internal functionality of space interpolating data value get methods.
			/// Valid process flags are: \e nearest, \e linear, \e bicubic
			/// This routine will only be successful provided the X vector is monotonically increasing or decreasing and no two values are equal, its result is undefined if applied to data not satisfying these conditions.
			/// \param x x co-ordinate
			/// \param It index in T-direction
			/// \param x_flag a valid process flag
			/// \param xa (returned) actual x value obtained via selected process
			/// \param v (returned) The data value
			void getDataItX(const long It, const float x, const processflag x_flag, float &xa, float &v) const;

			/// Common, internal functionality of time and space interpolating data value get methods.
			/// Valid process flags are: \e nearest, \e linear, \e bicubic
			/// This routine will only be successful provided the T and X vectors are monotonically increasing or decreasing and no two values are equal in each vector, its result is undefined if applied to data not satisfying these conditions.
			/// \param t time
			/// \param x x co-ordinate
			/// \param t_flag a valid process flag
			/// \param x_flag a valid process flag
			/// \param ta (returned) actual t value obtained via selected process
			/// \param xa (returned) actual x value obtained via selected process
			/// \param v (returned) The data value
			void getDataTX(const float t, const float x, const processflag t_flag, const processflag x_flag, float &ta, float &xa, float &v) const;
	};

};


#endif
