/*
 * PPF c++ wrapper
 *
 * Copyright (C) 2008 Alex Meakins
 *
 * 2008-04-09: Alex Meakins <alex.meakins@jet.uk>
 *             initial version
 *
 * This file is part of PPFc++.
 *
 * PPFc++ is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * PPFc++ is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser Public License for more details.
 *
 * You should have received a copy of the GNU Lesser Public License
 * along with PPFc++.  If not, see <http://www.gnu.org/licenses/>.
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <iostream>
#include <fstream>
#include <cstdlib>
#include <string>
#include "ppftypes.hpp"
#include "ppfdatatype.hpp"

using namespace std;

namespace ppf
{
	void datatype::resizeTX(int nt, int nx)
	{
		// check new sizes, any on axis set to zero sets both to zero
		if ((0 > nt) || (0 > nx)) throw ppf_error("Invalid number of elements specified.");
		if (0 == nt) nx = 0;
		else if (0 == nx) nt = 0;
		tvec.resize(nt);
		xvec.resize(nx);
		data.resize(nt);
		for (long i=0;i<data.size();++i) data[i].resize(nx);
	}
	
	void datatype::clear()
	{
		// set blank values
		tvec.clear();
		xvec.clear();
		data.clear();		
		xunit = "";
		tunit = "";
		dataunit = "";
		comment = ""; 
		userstatus = 0;	
		sysstatus = 0;	
	}

	// this method is currently quite crude (see ppf::write(filename) comment), there is no validation of the data atm
	void datatype::readLocal(const string filename)
	{
   		const string header = "PPFLDF";
		const short majorver = 1;
		const short minorver = 1;
		ifstream file;
		short tempsi;
		int tempi;
		char tempca[24];
		string tempst;

		// open file
		file.open(filename.c_str(), ios::binary);
		
		// check if file is open, report any errors
		if(!file.is_open()) throw ppf_error("Could not open specified local PPF file.");

		// read in header, check it is correct, report an error if it isn't
		file.read(tempca, 6);
		tempst.assign(tempca, 6);
		if (tempst != header)
		{
			file.close();
			throw ppf_error("Specified file is not a local PPF file.");
		}
		file.read((char*) &tempsi, sizeof(short));
		if (tempsi != majorver)
		{
			file.close();
			throw ppf_error("The version number of the specified local PPF file is not supported.");
		}
		file.read((char*) &tempsi, sizeof(short));
		if (tempsi != minorver)
		{
			file.close();
			throw ppf_error("The version number of the specified local PPF file is not supported.");
		}
		
		// read t unit string (8 chars)
		file.read(tempca, 8);
		tunit.assign(tempca, 8);

		// read x unit string (8 chars)
		file.read(tempca, 8);
		xunit.assign(tempca, 8);

		// read data unit string (8 chars)
		file.read(tempca, 8);
		dataunit.assign(tempca, 8);

		// read comment string (24 chars)
		file.read(tempca, 24);
		comment.assign(tempca, 24);

		// read status values
		file.read((char*) &tempi, sizeof(int));
		userstatus = tempi;
		file.read((char*) &tempi, sizeof(int));
		sysstatus = tempi;

		// read data t and x dimensions and set vector sizes
		file.read((char*) &tempi, sizeof(int)); // t dimension
		tvec.resize(tempi);
		data.resize(tempi);
		file.read((char*) &tempi, sizeof(int)); // x dimension
		xvec.resize(tempi);
		for(int i=0;i<data.size();i++) data[i].resize(tempi);

		// read t, x and data vectors
		for(int i=0;i<tvec.size();i++) file.read((char*) &tvec[i], sizeof(float));
		for(int i=0;i<xvec.size();i++) file.read((char*) &xvec[i], sizeof(float));
		for(int i=0;i<tvec.size();i++) for(int j=0;j<xvec.size();j++) file.read((char*) &data[i][j], sizeof(float));

		// close file
		file.close();
	}

	// this method is currently quite crude, it writes the X and T vectors even if they are not being used (ie all zeros)
	void datatype::writeLocal(const string filename)
	{
   		const string header = "PPFLDF";
		const short majorver = 1;
		const short minorver = 1;
   		ofstream file;
		int temp;

		// open file
		file.open(filename.c_str(), ios::binary);

		// check if file is open, report any errors
		if(!file.is_open()) throw ppf_error("The local PPF file could not be created.");

		// write header
		file.write(header.c_str(), 6);
		file.write((char*) &majorver, sizeof(short));
		file.write((char*) &minorver, sizeof(short));
		
		// write t unit string (8 chars)
		if (tunit.size() != 8) tunit.resize(8);
		file.write(tunit.c_str(), 8);

		// write x unit string (8 chars)
		if (xunit.size() != 8) xunit.resize(8);
		file.write(xunit.c_str(), 8);

		// write data unit string (8 chars)
		if (dataunit.size() != 8) dataunit.resize(8);
		file.write(dataunit.c_str(), 8);

		// write comment string (24 chars)
		if (comment.size() != 24) comment.resize(24);
		file.write(comment.c_str(), 24);

		// write status values
		file.write((char*) &userstatus, sizeof(int));
		file.write((char*) &sysstatus, sizeof(int));

		// write data t and x dimensions
		temp = tvec.size();
		file.write((char*) &temp, sizeof(int));
		temp = xvec.size();
		file.write((char*) &temp, sizeof(int));
		
		// write t, x and data vectors
		for(int i=0;i<tvec.size();i++) file.write((char*) &tvec[i], sizeof(float));
		for(int i=0;i<xvec.size();i++) file.write((char*) &xvec[i], sizeof(float));
   		for(int i=0;i<tvec.size();i++) for(int j=0;j<xvec.size();j++) file.write((char*) &data[i][j], sizeof(float));

		// close file
		file.close();
	}

	int datatype::getIt(const float t) const
	{
		int Ilower, Iupper;

		// find and return nearest index
		findNearest(tvec, 0, tvec.size()-1, t, Ilower, Iupper);
		if (Ilower == Iupper) return Ilower;
		if ((tvec[Iupper] - t) <= (t - tvec[Ilower])) return Iupper;
		else return Ilower;
	}

	int datatype::getIx(const float x) const
	{
		int Ilower, Iupper;

		// find and return nearest index
		findNearest(xvec, 0, xvec.size()-1, x, Ilower, Iupper);
		if (Ilower == Iupper) return Ilower;
		if ((xvec[Iupper] - x) <= (x - xvec[Ilower])) return Iupper;
		else return Ilower;
	}

	float datatype::getDataT(const float t, const processflag flag) const
	{
		float ta, v;

		getDataTIx(t, 0, flag, ta, v);
		return v;
	}

	coord datatype::getDataCT(const float t, const processflag flag) const
	{
		float ta, v;
		coord c;
		
		getDataTIx(t, 0, flag, ta, v);
		c.t = ta;
		c.x = xvec[0];
		c.d = v;
		return c;
	}

	float datatype::getDataX(const float x, const processflag flag) const
	{
		float xa, v;

		getDataItX(0, x, flag, xa, v);
		return v;
	}

	coord datatype::getDataCX(const float x, const processflag flag) const
	{
		float xa, v;
		coord c;

		getDataItX(0, x, flag, xa, v);
		c.t = tvec[0];
		c.x = xa;
		c.d = v;
		return c;
	}

	float datatype::getDataTIx(const float t, const long Ix, const processflag t_flag) const
	{
		float ta, v;

		getDataTIx(t, Ix, t_flag, ta, v);
		return v;
	}

	coord datatype::getDataCTIx(const float t, const long Ix, const processflag t_flag) const
	{
		float ta, v;
		coord c;

		getDataTIx(t, Ix, t_flag, ta, v);
		c.t = ta;
		c.x = xvec[Ix];
		c.d = v;
		return c;
	}

	float datatype::getDataItX(const long It, const float x, const processflag x_flag) const
	{
		float xa, v;

		getDataItX(It, x, x_flag, xa, v);
		return v;
	}

	coord datatype::getDataCItX(const long It, const float x, const processflag x_flag) const
	{
		float xa, v;
		coord c;

		getDataItX(It, x, x_flag, xa, v);
		c.t = tvec[It];
		c.x = xa;
		c.d = v;
		return c;
	}

	float datatype::getDataTX(const float t, const float x, const processflag t_flag, const processflag x_flag) const
	{
		float ta ,xa, v;

		getDataTX(t, x, t_flag, x_flag, ta, xa, v);
		return v;
	}

	coord datatype::getDataCTX(const float t, const float x, const processflag t_flag, const processflag x_flag) const
	{
		float ta ,xa, v;
		coord c;

		getDataTX(t, x, t_flag, x_flag, ta, xa, v);
		c.t = ta;
		c.x = xa;
		c.d = v;
		return c;
	}

	void datatype::setTUnit(const string s)
	{
		tunit = s;
		if (tunit.size() != 8) tunit.resize(8, ' ');
	}

	void datatype::setXUnit(const string s)
	{
		xunit = s;
		if (xunit.size() != 8) xunit.resize(8, ' ');
	}

	void datatype::setDataUnit(const string s)
	{
		dataunit = s;
		if (dataunit.size() != 8) dataunit.resize(8, ' ');	
	}

	void datatype::setComment(const string s)
	{
		comment = s;
		if (comment.size() != 24) comment.resize(24, ' ');
	}

	void datatype::findNearest(const vector<float> &v, const int Imin, const int Imax, const float f, int &Ilower, int &Iupper) const
	{
	// future work: add a linear prediction for the index of f from the gradient between the end points to give a mid point close to f = faster search, especialy for linear series!

		int t, b, m;	// top index, bottom index, middle index (for divide and conquer routine)

		// divide and conquer as appropriate
		if (v[Imin] < v[Imax]) // series is increasing
		{
			// do basic checks
			if (f <= v[Imin])
			{
				Ilower = Iupper = Imin;
				return;
			}
			if (f >= v[Imax])
			{
				Ilower = Iupper = Imax;
				return;
			}

			// answer is within specified range, so find it
			t = Imax;
			b = Imin;
			while(t != (b+1))
			{
				// find midpoint of range
				m = b + (t-b)/2; // note the integer divide

				// check if f was found, if not move the relevant side of the range inwards
				if (v[m] == f)
				{
					Ilower = Iupper = m;
					return;
				}
				if (v[m] < f) b = m;
				else t = m;
			}

			// return bounds
			Ilower = b;
			Iupper = t;
		}
		else // series is decreasing
		{
			// do basic checks
			if (f >= v[Imin])
			{
				Ilower = Iupper = Imin;
				return;
			}
			if (f <= v[Imax])
			{
				Ilower = Iupper = Imax;
				return;
			}
			
			// answer is within specified range, so find it
			t = Imax;
			b = Imin;
			while(t != (b+1))
			{
				// find midpoint of range
				m = b + (t-b)/2; // note the integer divide

				// check if f was found, if not move the relevant side of the range inwards
				if (v[m] == f)
				{
					Ilower = Iupper = m;
					return;
				}
				if (v[m] > f) b = m;
				else t = m;
			}

			// return bounds
			Ilower = t;
			Iupper = b;
		}
	}

	double datatype::interpolateCubic(const double p1, const double g1, const double v1, const double p2, const double g2, const double v2, const double pt) const
	{
		// cache common calculations
		double p1s = p1*p1;
		double p2s = p2*p2;
		double p1c = p1*p1*p1;
		double pts = pt*pt;

		// calc spline coefficients
		double A = ( 2*(v1-v2) + (g1 + g2) * (p2 - p1) ) / ( 3*(p1s*p2 - p2s*p1) - p1c + p2*p2*p2 );
		double B = (g1 - g2 + 3*A*(p2s - p1s)) / (2*(p1-p2));
		double C = g1 - 3*A*p1s - 2*B*p1;
		double D = v1 - A*p1c - B*p1s - C*p1;

		// evaluate spline
		return A*pt*pts + B*pts + C*pt + D;
	}

	void datatype::getDataTIx(const float t, const long Ix, const processflag t_flag, float &ta, float &v) const
	{
		int Ilower, Iupper;

		if(t_flag == ppf::nearest)    // FIND NEAREST POINT
		{
			// find nearest index and return its corresponding data value
			findNearest(tvec, 0, tvec.size()-1, t, Ilower, Iupper);
			if (Ilower == Iupper)
			{
				ta = tvec[Ilower];
				v = data[Ilower][Ix];
			}
			else if ((tvec[Iupper] - t) <= (t - tvec[Ilower]))
			{
				ta = tvec[Iupper];
				v = data[Iupper][Ix];
			}
			else
			{
				ta = tvec[Ilower];
				v = data[Ilower][Ix];
			}
		}
		else if(t_flag == ppf::linear)    // LINEAR INTERPOLATION
		{
			// linearly interpolate if appropriate (not at end of array) and return values
			findNearest(tvec, 0, tvec.size()-1, t, Ilower, Iupper);
			if (Ilower == Iupper)
			{
				// value has been exactly identified so return it
				ta = tvec[Ilower];
				v = data[Ilower][Ix];
			}
			else
			{
				// interpolate
				ta = t;
				v = interpolateLinear(tvec[Ilower], data[Ilower][Ix], tvec[Iupper], data[Iupper][Ix], t);
			}
		}
		else if(t_flag == ppf::cubic)    // CUBIC INTERPOLATION
		{
			// cubic interpolate if appropriate (not at end of array) and return values
			findNearest(tvec, 0, tvec.size()-1, t, Ilower, Iupper);
			if (Ilower == Iupper)
			{
				// value has been exactly identified so return it
				ta = tvec[Ilower];
				v = data[Ilower][Ix];
			}
			else
			{
				double gl, gu;

				// calculate gradients, take care at end points
				if (Ilower == 0) gl = (data[Ilower+1][Ix]-data[Ilower][Ix]) / (tvec[Ilower+1] - tvec[Ilower]);
				else if (Ilower == (tvec.size()-1)) gl = (data[Ilower][Ix]-data[Ilower-1][Ix]) / (tvec[Ilower] - tvec[Ilower-1]);
				else gl = 0.5*(data[Ilower+1][Ix]-data[Ilower][Ix]) / (tvec[Ilower+1] - tvec[Ilower]) + 0.5*(data[Ilower][Ix]-data[Ilower-1][Ix]) / (tvec[Ilower] - tvec[Ilower-1]);
				
				if (Iupper == 0) gu = (data[Iupper+1][Ix]-data[Iupper][Ix]) / (tvec[Iupper+1] - tvec[Iupper]);
				else if (Iupper == (tvec.size()-1)) gu = (data[Iupper][Ix]-data[Iupper-1][Ix]) / (tvec[Iupper] - tvec[Iupper-1]);
				else gu = 0.5*(data[Iupper+1][Ix]-data[Iupper][Ix]) / (tvec[Iupper+1] - tvec[Iupper]) + 0.5*(data[Iupper][Ix]-data[Iupper-1][Ix]) / (tvec[Iupper] - tvec[Iupper-1]);

				// interpolate
				ta = t;
				v = interpolateCubic(tvec[Ilower], gl, data[Ilower][Ix], tvec[Iupper], gu, data[Iupper][Ix], t);
			}
		}
		else throw ppf_error("Invalid process flag.");
	}

	void datatype::getDataItX(const long It, const float x, const processflag x_flag, float &xa, float &v) const
	{
		int Ilower, Iupper;

		if(x_flag == ppf::nearest)    // FIND NEAREST POINT
		{
			// find nearest index and return its corresponding data value
			findNearest(xvec, 0, xvec.size()-1, x, Ilower, Iupper);
			if (Ilower == Iupper)
			{
				xa = xvec[Ilower];
				v = data[It][Ilower];
			}
			else if ((xvec[Iupper] - x) <= (x - xvec[Ilower]))
			{
				xa = xvec[Iupper];
				v = data[It][Iupper];
			}
			else
			{
				xa = xvec[Ilower];
				v = data[It][Ilower];
			}
		}
		else if(x_flag == ppf::linear)    // LINEAR INTERPOLATION
		{
			// linearly interpolate if appropriate (not at end of array) and return values
			findNearest(xvec, 0, xvec.size()-1, x, Ilower, Iupper);
			if (Ilower == Iupper)
			{
				// value has been exactly identified so return it
				xa = xvec[Ilower];
				v = data[It][Ilower];
			}
			else
			{
				// interpolate
				xa = x;
				v = interpolateLinear(xvec[Ilower], data[It][Ilower], xvec[Iupper], data[It][Iupper], x);
			}
		}
		else if(x_flag == ppf::cubic)    // CUBIC INTERPOLATION
		{
			// cubic interpolate if appropriate (not at end of array) and return values
			findNearest(xvec, 0, xvec.size()-1, x, Ilower, Iupper);
			if (Ilower == Iupper)
			{
				// value has been exactly identified so return it
				xa = xvec[Ilower];
				v = data[It][Ilower];
			}
			else
			{
				double gl, gu;

				// calculate gradients, take care at end points
				if (Ilower == 0) gl = (data[It][Ilower+1]-data[It][Ilower]) / (xvec[Ilower+1] - xvec[Ilower]);
				else if (Ilower == (xvec.size()-1)) gl = (data[It][Ilower] - data[It][Ilower-1]) / (xvec[Ilower] - xvec[Ilower-1]);
				else gl = 0.5*(data[It][Ilower+1]-data[It][Ilower]) / (xvec[Ilower+1] - xvec[Ilower]) + 0.5*(data[It][Ilower] - data[It][Ilower-1]) / (xvec[Ilower] - xvec[Ilower-1]);
				
				if (Iupper == 0) gu = (data[It][Iupper+1]-data[It][Iupper]) / (xvec[Iupper+1] - xvec[Iupper]);
				else if (Iupper == (xvec.size()-1)) gu = (data[It][Iupper] - data[It][Iupper-1]) / (xvec[Iupper] - xvec[Iupper-1]);
				else gu = 0.5*(data[It][Iupper+1]-data[It][Iupper]) / (xvec[Iupper+1] - xvec[Iupper]) + 0.5*(data[It][Iupper] - data[It][Iupper-1]) / (xvec[Iupper] - xvec[Iupper-1]);

				// interpolate
				xa = x;
				v = interpolateCubic(xvec[Ilower], gl, data[It][Ilower], xvec[Iupper], gu, data[It][Iupper], x);
			}
		}
		else throw ppf_error("Invalid process flag.");
	}

	void datatype::getDataTX(const float t, const float x, const processflag t_flag, const processflag x_flag, float &ta, float &xa, float &v) const
	{
		int Ilower, Iupper;

		if(x_flag == ppf::nearest)        // FIND NEAREST POINT (X)
		{
			// find nearest X indicies
			findNearest(xvec, 0, xvec.size()-1, x, Ilower, Iupper);
			if (Ilower == Iupper)
			{
				// set xa, call routine to get ta and v as per process flag
				xa = xvec[Ilower];
				getDataTIx(t, Ilower, t_flag, ta, v);
			}
			else if ((xvec[Iupper] - x) <= (x - xvec[Ilower]))
			{
				// upper value is nearer, call routine for ta and v
				xa = xvec[Iupper];
				getDataTIx(t, Iupper, t_flag, ta, v);
			}
			else
			{
				// lower value is nearer, call routine for ta and v
				xa = xvec[Ilower];
				getDataTIx(t, Ilower, t_flag, ta, v);
			}
		}
		else if (x_flag == ppf::linear)   // LINEAR INTERPOLATION (X)
		{
			// linearly interpolate if appropriate (not at end of array) and return values
			findNearest(xvec, 0, xvec.size()-1, x, Ilower, Iupper);
			if (Ilower == Iupper)
			{
				// value has been exactly identified so return it, call routine to get ta and v
				xa = xvec[Ilower];
				getDataTIx(t, Ilower, t_flag, ta, v);
			}
			else
			{
				float nt, vl, vu;

				// collect the processed t values
				getDataTIx(t, Ilower, t_flag, nt, vl);
				getDataTIx(t, Iupper, t_flag, nt, vu);

				// interpolate and return data
				xa = x;
				ta = nt;
				v = interpolateLinear(xvec[Ilower], vl, xvec[Iupper], vu, x);
			}
		}
		else if(x_flag == ppf::cubic)    // CUBIC INTERPOLATION (X)
		{
			// cubic interpolate if appropriate (not at end of array) and return values
			findNearest(xvec, 0, xvec.size()-1, x, Ilower, Iupper);
			if (Ilower == Iupper)
			{
				// value has been exactly identified so return it, call routine to get ta and v
				xa = xvec[Ilower];
				getDataTIx(t, Ilower, t_flag, ta, v);
			}
			else
			{
				float nt, vll, vl, vu, vuu;
				double gl, gu;

				// collect the processed t values, check series order
				if (Ilower < Iupper)	
				{
					// positive series
					getDataTIx(t, Ilower-1, t_flag, nt, vll);
					getDataTIx(t, Ilower, t_flag, nt, vl);
					getDataTIx(t, Iupper, t_flag, nt, vu);
					getDataTIx(t, Iupper+1, t_flag, nt, vuu);

					// calculate gradients, take care at end points
					if (Ilower == 0) gl = (vu - vl) / (xvec[Iupper] - xvec[Ilower]);
					else if (Ilower == (xvec.size()-1)) gl = (vl - vll) / (xvec[Ilower] - xvec[Ilower-1]);
					else gl = 0.5*(vu - vl) / (xvec[Iupper] - xvec[Ilower]) + 0.5*(vl - vll) / (xvec[Ilower] - xvec[Ilower-1]);
					
					if (Iupper == 0) gu = (vuu - vu) / (xvec[Iupper+1] - xvec[Iupper]);
					else if (Iupper == (xvec.size()-1)) gu = (vu - vl) / (xvec[Iupper] - xvec[Ilower]);
					else gu = 0.5*(vuu - vu) / (xvec[Iupper+1] - xvec[Iupper]) + 0.5*(vu - vl) / (xvec[Iupper] - xvec[Ilower]);
				}
				else	
				{
					// negative series
					getDataTIx(t, Ilower+1, t_flag, nt, vll);
					getDataTIx(t, Ilower, t_flag, nt, vl);
					getDataTIx(t, Iupper, t_flag, nt, vu);
					getDataTIx(t, Iupper-1, t_flag, nt, vuu);
	
					// calculate gradients, take care at end points
					if (Ilower == 0) gl = (vll - vl) / (xvec[Ilower+1] - xvec[Ilower]);
					else if (Ilower == (xvec.size()-1)) gl = (vl - vu) / (xvec[Ilower] - xvec[Iupper]);
					else gl = 0.5*(vll - vl) / (xvec[Ilower+1] - xvec[Ilower]) + 0.5*(vl - vu) / (xvec[Ilower] - xvec[Iupper]);
				
					if (Iupper == 0) gu = (vl - vu) / (xvec[Ilower] - xvec[Iupper]);
					else if (Iupper == (xvec.size()-1)) gu = (vu - vuu) / (xvec[Iupper] - xvec[Iupper-1]);
					else gu = 0.5*(vl - vu) / (xvec[Ilower] - xvec[Iupper]) + 0.5*(vu - vuu) / (xvec[Iupper] - xvec[Iupper-1]);
				}

				// interpolate
				xa = x;
				ta = nt;
				v = interpolateCubic(xvec[Ilower], gl, vl, xvec[Iupper], gu, vu, x);
			}
		}
		else throw ppf_error("Invalid process flag.");
	}
};
