/*
 * PPF c++ wrapper
 *
 * Copyright (C) 2008 Alex Meakins
 *
 * 2008-04-09: Alex Meakins <alex.meakins@jet.uk>
 *             initial version
 *
 * This file is part of PPFc++.
 *
 * PPFc++ is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * PPFc++ is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser Public License for more details.
 *
 * You should have received a copy of the GNU Lesser Public License
 * along with PPFc++.  If not, see <http://www.gnu.org/licenses/>.
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <iostream>
#include <fstream>
#include <cstdlib>
#include <string>
#include "ppftypes.hpp"
#include "ppfsystem.hpp"
#include "ppfdatatype.hpp"
extern "C"{
#include "ppf.h"
};

using namespace std;

namespace ppf
{
	// global variables --------------------------------------------------------
	system ppfdb;

	// class methods -----------------------------------------------------------
	void system::connect() const
	{
		connect("JETPPF", ppf::read);
	}

	void system::connect(string user) const
	{
		connect(user, ppf::read);
	}
	
	void system::connect(string user, const accessflag flag) const
	{
		char caUserName[9];
		char cR[2] = "R";
		char cW[2] = "W";

		// check access flags are valid
		if (!(flag & ppf::read) && !(flag & ppf::write))
		{
			// throw error
			throw ppf_error("Invalid access flags specified in call to connect().");
		}

		// make sure user name is no more then 8 characters
		if (user.size() > 8) user.erase(8);
		strcpy(caUserName, user.c_str());

		// call PPFUID for reading and/or writing as requested via access flags
		if (flag & ppf::read) PPFUID(caUserName, cR, 8,1);
		if (flag & ppf::write) PPFUID(caUserName, cW, 8,1);
	}

	void system::read(datatype &d, const long shot, string DDA, string dtype) const
	{
		read(d, shot, 0, DDA, dtype, ppf::xvector | ppf::tvector);
	}
	
	void system::read(datatype &d, const long shot, string DDA, string dtype, const ioflag flag) const
	{
		read(d, shot, 0, DDA, dtype, flag);
	}
	
	void system::read(datatype &d, const long shot, const long sequence, string DDA, string dtype) const
	{
		read(d, shot, sequence, DDA, dtype,  ppf::xvector | ppf::tvector);
	}

	void system::read(datatype &d, const long shot, const long sequence, string DDA, string dtype, const ioflag flag) const
	{
   		int iError = 0;
		int iShot = shot;
		int iSequence = sequence;
		int iLun = 0;
		int nX = 0;
		int nT = 0;
		int nData = 0;
		int iReadData[13];
		int iWriteData[13];
		char caDDA[5];
		char caDataType[5];
		char caHeaderData[60];
		float *fData = 0;
		float *fX = 0;
		float *fT = 0;
		bool usexvector;
		bool usetvector;

		// initialise the ppf system for reading
		PPFGO(&iShot, &iSequence , &iLun , &iError);

		// process any errors
		if (iError) throw ppf_error("PPFGO", iError);

		// check size and copy DDA and dtype strings to char arrays (can't use .c_str() as this returns a const char*, PPFGET need a plain char*)
		DDA.resize(4, ' ');
		strcpy(caDDA, DDA.c_str());
		dtype.resize(4, ' ');
		strcpy(caDataType, dtype.c_str());

		// set variables as required to determine data size using PPFGET
		iReadData[0] = 0;	// NX1
		iReadData[1] = 0;	// NX2
		iReadData[2] = 0;	// NT1
		iReadData[3] = 0;	// NT2
		iReadData[4] = 0;	// NDMAX, set to zero so no data is returned
		iReadData[5] = 0;	// NXMAX, set to zero so no data is returned
		iReadData[6] = 0;	// NTMAX, set to zero so no data is returned
		iReadData[7] = -1;  // don't want x-vector at this stage
		iReadData[8] = -1;  // don't want x-vector at this stage
		iReadData[9] = iReadData[10] = iReadData[11] = iReadData[12] = 0;   // Not used so set to zero

		// open datatype without loading data to determine array size required
		PPFGET(&iShot, caDDA, caDataType, iReadData, caHeaderData, iWriteData, fData, fX, fT, &iError, 5, 5, 60);

		// check for errors
		if (iError) throw ppf_error("PPFGET", iError);

		// get data array sizes
		nX = iWriteData[5];
		nT = iWriteData[6];
		nData = nX * nT;

		// if nData, nX or nT is < 1 throw an error
		if ((nX < 1) || (nT < 1) || (nData < 1)) throw ppf_error("Invalid data ranges returned for datatype.");

		// check if we are using the t or x vectors and set class flags accordingly
		usexvector = !(flag & ppf::noxvector);
		usetvector = !(flag & ppf::notvector);

		// create arrays to hold data as necessary
		fData = new float[nData];
		if (usexvector) fX = new float[nX];    // only create array if noxvector flag is not set
		if (usetvector) fT = new float[nT];    // only create array if notvector flag is not set

		// update iReadData variables as required to read data with PPFGET
		iReadData[4] = nData;	// NDMAX
		iReadData[5] = nX;		// NXMAX
		iReadData[6] = nT;		// NTMAX
		if (usexvector) iReadData[7] = 0;	// was set to -1 in previous definition so re-enable if required
		if (usetvector) iReadData[8] = 0;	// was set to -1 in previous definition so re-enable if required

		// open datatype without loading data to determine array size required
		PPFGET(&iShot, caDDA, caDataType, iReadData, caHeaderData, iWriteData, fData, fX, fT, &iError, 5, 5, 60);

		// check for errors
		if (iError)
		{
			if (fX) delete [] fX;
			if (fT) delete [] fT;
			delete [] fData;
			throw ppf_error("PPFGET", iError);
		}

		// resise x, t and data vectors
		d.data.resize(nT);
		for (int i=0;i<nT;i++) d.data[i].resize(nX);
		d.xvec.resize(nX);
		d.tvec.resize(nT);

		// copy data to class vectors (using an array then copying into a vector is inefficient - must rewrite class in the future)
		for(int i=0;i<nT;i++) for(int j=0;j<nX;j++) d.data[i][j] = fData[i*nX + j];
		if (usexvector) for(int i=0;i<nX;i++) d.xvec[i] = fX[i];
		else for(int i=0;i<nX;i++) d.xvec[i] = 0;
		if (usetvector) for(int i=0;i<nT;i++) d.tvec[i] = fT[i];
		else for(int i=0;i<nT;i++) d.tvec[i] = 0;

		// place header information in appropriate class strings
		d.dataunit.assign(&caHeaderData[0], 8);
		d.xunit.assign(&caHeaderData[8], 8);
		d.tunit.assign(&caHeaderData[16], 8);
		d.comment.assign(&caHeaderData[36], 24);

		// set status values
		d.userstatus = iWriteData[9];
		d.sysstatus = iWriteData[11];

		// delete arrays
		if (fX) delete [] fX;
		if (fT) delete [] fT;
		delete [] fData;
	}

	
	void system::create(const long shot)
	{
		create(shot, 0, "");
	}

	void system::create(const long shot, const long status)
	{
		create(shot, status, "");
	}
	
	void system::create(const long shot, const string comment)
	{
		create(shot, 0, comment);
	}
	
	void system::create(long shot, long status, string comment)
	{
		int iShot = shot;
		int iSequence = 0;
		int iLun = 0;
		int iNRMax = 0;
		int iInputData[10];
		char *caComment;
   		int iError = 0;
   		int temp;
   		
		// pad out comment to make sure length is a mutiple of 4
		temp = comment.size() / 4;
		if ((temp*4) < comment.size()) ++temp;
		comment.resize(temp*4,' ');

		// copy comment and set iInputData[9]
		iInputData[9] = temp;
		caComment = new char[(temp*4)+1];
		strcpy(caComment, comment.c_str());

		// set remaining iInputData values
		iInputData[0] = shot;
		iInputData[1] = 0;  // date set blank
		iInputData[2] = 0;  // time set blank
		iInputData[3] = status;
		iInputData[4] = 0;
		iInputData[5] = 0;
		iInputData[6] = 0;
		iInputData[7] = 0;
		iInputData[8] = 0;

		// create new ppf
		PPFOPN(&iShot, &iSequence, &iLun, &iNRMax, iInputData, caComment, &iError, (temp*4)+1);
		
		// report any error
		if (iError)
		{
	         	delete [] caComment;
			throw ppf_error("PPFOPN", iError);
		}

		// set openshot to the correct shot number for thw new ppf
		openshot = shot;

		// reset X/T vector reference variables to -1 (so that if the user tried to write a the first datatype requesting reuse of vectors the writing of vectors will be forced)
		lastXRef = -1;		
		lastTRef = -1;

		// clean up
   		delete [] caComment;
	}

	void system::write(datatype &d, string DDA, string dtype)
	{
		write(d, DDA, dtype, ppf::xvector | ppf::tvector);
	}

	void system::write(datatype &d, string DDA, string dtype, const ioflag flag)
	{
  		int iError = 0;
		int iShot = openshot;
		int iReadData[13];
		int iWriteData[13];
		char caDDA[5];
		char caDataType[5];
		char caHeaderData[60];
		float *fData = 0;
		float *fX = 0;
		float *fT = 0;
		const string storagetype = "F   ";

		// check size and copy DDA and dtype strings to char arrays (can't use .c_str() as this returns a const char*, PPFGET need a plain char*)
		DDA.resize(4, ' ');
		strcpy(caDDA, DDA.c_str());
		dtype.resize(4, ' ');
		strcpy(caDataType, dtype.c_str());
					
		// set iReadData, vector references determined by flags		
		iReadData[0] = 0;
		iReadData[1] = 0;
		iReadData[2] = 0;
		iReadData[3] = 0;
		iReadData[4] = 0;
		iReadData[5] = d.sizeX();
		iReadData[6] = d.sizeT();
		if (flag & ppf::noxvector) iReadData[7] = 0;
		else if (flag & ppf::lastxvector) iReadData[7] = lastXRef;
		else iReadData[7] = -1;
		if (flag & ppf::notvector) iReadData[8] = 0;
		else if (flag & ppf::lasttvector) iReadData[8] = lastTRef;
		else iReadData[8] = -1;
		iReadData[9] = d.getUserStatus();
		iReadData[10] = d.getSysStatus();
		iReadData[11] = 0;		
		iReadData[12] = 0;

		// set caHeaderData 
		d.dataunit.copy(&caHeaderData[0],8,0);
		d.xunit.copy(&caHeaderData[8],8,0);
		d.tunit.copy(&caHeaderData[16],8,0);
		storagetype.copy(&caHeaderData[24],4,0);
		storagetype.copy(&caHeaderData[28],4,0);
		storagetype.copy(&caHeaderData[32],4,0);		
		d.comment.copy(&caHeaderData[36],24,0);

		// create arrays to hold data (must allocate space even if it isn't used - PPFWRI is fickle and stupid)
		fData = new float[d.sizeX() * d.sizeT()];
		fX = new float[d.sizeX()];
		fT = new float[d.sizeT()];

		// copy data from datatype to temporary arrays (this will be removed once datatype is rewritten to remove the use of STL vectors, or direct connections to the PPF system can be used)
		for(int i=0;i<d.sizeT();i++) for(int j=0;j<d.sizeX();j++) fData[i*d.sizeX() + j] = d.data[i][j];
		if (iReadData[7] == -1) for(int i=0;i<d.sizeX();i++) fX[i] = d.xvec[i];
		if (iReadData[8] == -1) for(int i=0;i<d.sizeT();i++) fT[i] = d.tvec[i];

		// write
		PPFWRI(&iShot, caDDA, caDataType, iReadData, caHeaderData, iWriteData, fData, fX, fT, &iError, 5, 5, 60);

		// check for errors
		if (iError)
		{
			delete [] fData;
			delete [] fX;
			delete [] fT;
			throw ppf_error("PPFWRI", iError);
		}

		// set last X/T vector references if new X/T vectors written
		if (iReadData[7] == -1) lastXRef = iWriteData[7];
		if (iReadData[8] == -1) lastTRef = iWriteData[8];

		// clean up
		delete [] fData;
		delete [] fX;
		delete [] fT;
	}

	void system::closeDDA() const
	{
		closeDDA(0, "");
	}
	
	void system::closeDDA(const int status) const
	{
		closeDDA(status, "");
	}
	
	void system::closeDDA(string comment) const
	{
		closeDDA(0, comment);
	}
	
	void system::closeDDA(const int status, string comment) const
	{
		int iStatus = status;
		int iCommentWords = 0;
		char *caComment;
		int iError = 0;

		// pad out comment to make sure length is a mutiple of 4
		iCommentWords = comment.size() / 4;
		if ((iCommentWords*4) < comment.size()) ++iCommentWords;
		comment.resize(iCommentWords*4,' ');

		// copy comment
		caComment = new char[(iCommentWords*4)+1];
		strcpy(caComment, comment.c_str());
		
		// close DDA
		DDACLO(&iStatus, &iCommentWords, caComment, &iError, (iCommentWords*4)+1);
		
		// check for errors
		if (iError)
		{
			delete [] caComment;
			throw ppf_error("DDACLO", iError);
		}
		// clean up
		delete [] caComment;

	}

	void system::abort() const
	{
		int iError;
		
		// abort ppf
		PPFABO(&iError);

		// check for errors
		if (iError) throw ppf_error("PPFABO", iError);
	}
	
	int system::commit() const
	{
		return commit("",0);
	}

	int system::commit(string program, int version) const
	{
		int iShot = openshot;
		int iSequence = 0;
		int iLun = 0;
		char caProgramName[9];
		int iVersion = version;
		int iError = 0;
		
		// check program name is within 8 chars, pad with spaces if not, copy to buffer
		program.resize(8,' ');
		strcpy(caProgramName, program.c_str());
		
		// close ppf
		PPFCLO(&iShot, &iSequence, &iLun, caProgramName, &iVersion, &iError, 9);
		
		// check for errors
		if (iError) throw ppf_error("PPFCLO", iError);

		// return sequence number
		return iSequence;
	}
};
